#include <stdio.h>
#include <stdlib.h>

struct cell {
  int value;
  struct cell *prev, *next;
};

struct list {
  struct cell *first, *last;
};

void push_front(struct list *l, int value)
{
  struct cell *c;

  c = malloc(sizeof(struct cell));
  c->value = value;
  c->next = l->first;
  c->prev = NULL;

  if (l->first != NULL)
    l->first->prev = c;

  l->first = c;
  if (l->last == NULL)
    l->last = c;
}

void print(struct list *l)
{
  struct cell *c;

  printf("[");
  for (c = l->first; c != NULL; c = c->next) {
    printf("%d", c->value);
    if (c->next != NULL)
      printf(", ");
  }
  printf("]\n");
}

void push_back(struct list *l, int value)
{
  struct cell *c;

  c = malloc(sizeof(struct cell));
  c->value = value;
  c->prev = l->last;
  c->next = NULL;

  if (l->last != NULL)
    l->last->next = c;

  l->last = c;
  if (l->first == NULL)
    l->first = c;
}

int pop_front(struct list *l)
{
  struct cell *c;
  int value;

  if (l->first == NULL) {
    printf("pop_front d'une liste vide.\n");
    abort();
  }

  c = l->first;
  value = c->value;
  l->first = c->next;
  free(c);

  if (l->first != NULL)
    l->first->prev = NULL;
  else
    l->last = NULL;

  return value;
}

int pop_back(struct list *l)
{
  struct cell *c;
  int value;

  if (l->last == NULL) {
    printf("pop_back d'une liste vide.\n");
    abort();
  }

  c = l->last;
  value = c->value;
  l->last = c->prev;
  free(c);

  if (l->last != NULL)
    l->last->next = NULL;
  else
    l->first = NULL;

  return value;
}

void reverse(struct list *l)
{
  struct cell *c, *prev, *next;

  l->last = l->first;
  for (c = l->first, prev = NULL; c != NULL; c = next) {
    next = c->next;
    c->prev = next;
    c->next = prev;
    prev = c;
  }
  l->first = prev;
}

void append(struct list *l1, struct list *l2, struct list *l)
{
  l->first = (l1->first != NULL) ? l1->first : l2->first;
  l->last = (l2->last != NULL) ? l2->last : l1->last;

  if ((l1->last != NULL) && (l2->first != NULL)) {
    l1->last->next = l2->first;
    l2->first->prev = l1->last;
  }
}

int main()
{
  struct list l1 = { NULL, NULL };
  struct list l2 = { NULL, NULL };
  struct list l;

  push_front(&l1, 4);
  push_front(&l1, 6);
  push_front(&l1, 16);

  push_back(&l2, 51);
  push_back(&l2, 42);

  print(&l1);
  print(&l2);
  append(&l1, &l2, &l);
  print(&l);
  reverse(&l);
  print(&l);

  return 0;
}
